@@ -10,7 +10,7 @@ import pandas as pd
 from ._prompt_builder import PromptBuilder
 from ._test_tree_browser import TestTreeBrowser, is_subtopic
 from ._model import Model
-from ._topic_model import TopicModel
+from ._topic_model import TopicLabelingModel, TopicMembershipModel
 import adatest
 from pathlib import Path
 import torch
@@ -36,7 +36,7 @@ class TestTree():
     webserver. A TestTree object also conforms to most of the standard pandas DataFrame API.
     """
 
-    def __init__(self, tests=None, index=None, compute_embeddings=False, **kwargs):
+    def __init__(self, tests=None, index=None, compute_embeddings=False, ensure_topic_markers=True, **kwargs):
         """ Create a new test tree.
 
         Parameters
@@ -118,6 +118,34 @@ class TestTree():
             self._tests["labeler"] = ["imputed" for _ in range(self._tests.shape[0])]
 
         # ensure that all topics have a topic_marker entry
+        if ensure_topic_markers:
+            self.ensure_topic_markers()
+
+        # drop any duplicate index values
+        self._tests = self._tests.groupby(level=0).first()
+
+        # drop any duplicate rows
+        self._tests.drop_duplicates(["topic", "input", "output", "labeler"], inplace=True)
+
+        # put the columns in a consistent order
+        self._tests = self._tests[column_names + [c for c in self._tests.columns if c not in column_names]]
+
+        if compute_embeddings:
+            self._cache_embeddings()
+
+        # replace any invalid topics with the empty string
+        for i, row in self._tests.iterrows():
+            if not isinstance(row.topic, str) or not row.topic.startswith("/"):
+                self._tests.loc[i, "topic"] = ""
+
+        self._topic_labeling_models = {}
+        self._topic_membership_models = {}
+
+        # # keep track of our original state
+        # if self.auto_save:
+        #     self._last_saved_tests = self._tests.copy()
+
+    def ensure_topic_markers(self):
         marked_topics = {t: True for t in set(self._tests.loc[self._tests["label"] == "topic_marker"]["topic"])}
         for topic in set(self._tests["topic"]):
             parts = topic.split("/")
@@ -134,71 +162,6 @@ class TestTree():
                     }
                     marked_topics[parent_topic] = True
 
-        # drop any duplicate index values
-        self._tests = self._tests.groupby(level=0).first()
-
-        # drop any duplicate rows
-        self._tests.drop_duplicates(["topic", "input", "output", "labeler"], inplace=True)
-
-        # put the columns in a consistent order
-        self._tests = self._tests[column_names + [c for c in self._tests.columns if c not in column_names]]
-
-        if compute_embeddings:
-            self.compute_embeddings()
-
-        self._topic_models = {}
-
-        # # keep track of our original state
-        # if self.auto_save:
-        #     self._last_saved_tests = self._tests.copy()
-
-    def compute_embeddings(self):
-        """ Compute the embeddings for all test cases in the test tree.
-
-        TODO: Shared logic with TestTreeBrowser._compute_embeddings_and_scores. Refactor! (SML: I think it will live here now)
-        """
-
-        # see what new embeddings we need to compute
-        all_strings = []
-        for id, test in self._tests.iterrows():
-            if test.label == "topic_marker":
-                parts = test.topic.rsplit("/", 1)
-                str = parts[1] if len(parts) == 2 else ""
-                all_strings.append(str)
-            else:
-                for str in [test.input, test.output]:
-                    all_strings.append(str)
-
-        # suggestions topics don't have topic markers so we check for them separately
-        all_strings.append("__suggestions__")
-        
-        # we don't use the output of the embedding, just do this to get the embeddings cached
-        adatest.embed(all_strings)
-
-        # compute the embeddings
-        # if len(new_strings) > 0:
-        #     string_embeddings = adatest.embed(new_strings).cpu()
-        #     for i,s in enumerate(new_strings):
-        #         adatest._embedding_cache[s] = string_embeddings[i]
-
-        # compute the embeddings
-        # if len(new_embedding_ids) > 0:
-        #     inputs = []
-        #     outputs = []
-        #     for id in new_embedding_ids:
-        #         if self._tests.loc[id, "label"] == "topic_marker":
-        #             parts = self._tests.loc[id, "topic"].rsplit("/", 1)
-        #             inputs.append(parts[1] if len(parts) == 2 else "")
-        #             outputs.append("")
-        #         else:
-        #             inputs.append(str(self._tests.loc[id, "input"]))
-        #             outputs.append(str(self._tests.loc[id, "output"]))
-        #     new_strings = list(set(new_strings))
-        #     string_embeddings = adatest.embedding_model.encode(strings, convert_to_tensor=True, show_progress_bar=False).cpu()
-        #     for i,s in enumerate(strings):
-        #         adatest._embedding_cache[s] = string_embeddings[i]
-
-
     def __getitem__(self, key):
         """ TestSets act just like a DataFrame when sliced. """
         subset = self._tests[key]
@@ -241,6 +204,9 @@ class TestTree():
     @property
     def copy(self):
         return self._tests.copy
+    @property
+    def sort_values(self):
+        return self._tests.sort_values
     
     # NOTE: Can't delegate to df.append as it is depricated in favor of pd.concat, which we can't use due to type checks 
     def append(self, test_tree, axis=0):
@@ -248,9 +214,15 @@ class TestTree():
             self._tests = pd.concat([self._tests, test_tree], axis=axis)
         elif isinstance(test_tree, TestTree):
             self._tests = pd.concat([self._tests, test_tree._tests], axis=axis)
+        elif isinstance(test_tree, dict):
+            # check if the values are strings or lists of strings
+            if any([isinstance(v, str) for v in test_tree.values()]):
+                self._tests = pd.concat([self._tests, pd.DataFrame({k: [test_tree[k]] for k in test_tree}, index=[uuid.uuid4().hex])], axis=axis)
+            else:
+                self._tests = pd.concat([self._tests, pd.DataFrame(test_tree)], axis=axis)
 
-        self.deduplicate()
-        self.compute_embeddings()
+        #self.deduplicate()
+        #self.compute_embeddings()
         return None # TODO: Rethink append logic -- return copy vs. in place update?
 
     def __len__(self):
@@ -275,9 +247,9 @@ class TestTree():
         ids = [id for id, test in self._tests.iterrows() if is_subtopic(topic, test.topic)]
         return self.loc[ids]
 
-    def __call__(self, scorer=None, generator=None, auto_save=False, user="anonymous", recompute_scores=False, drop_inactive_score_columns=False,
-                 max_suggestions=100, suggestion_thread_budget=0.5, prompt_builder=PromptBuilder(), active_generator="default", starting_path="",
-                 score_filter=-1e10, topic_model_scale=0): # TODO: remove active_generator and replace with the ability to set the generator?
+    def adapt(self, scorer=None, generator=adatest.generators.OpenAI(), auto_save=False, user="anonymous", recompute_scores=False, drop_inactive_score_columns=False,
+              max_suggestions=100, suggestion_thread_budget=0.5, prompt_builder=PromptBuilder(), active_generator="default", starting_path="",
+              score_filter=-1e10, topic_model_scale=0): # TODO: remove active_generator and replace with the ability to set the generator?
         """ Apply this test tree to a scorer/model and browse/edit the tests to adapt them to the target model.
 
         Applying a test tree to a target model (wrapped by a scorer) creates a TestTreeBrowser object that can be used to
@@ -367,66 +339,96 @@ class TestTree():
                 already_seen[k] = True
         self._tests = self._tests.drop(drop_ids, axis=0)
 
+    def _cache_embeddings(self):
+        """ Pre-compute the embeddings for all test cases in the test tree.
+        """
+
+        # see what new embeddings we need to compute
+        all_strings = []
+        for id, test in self._tests.iterrows():
+            if test.label == "topic_marker":
+                parts = test.topic.rsplit("/", 1)
+                str = parts[1] if len(parts) == 2 else ""
+                all_strings.append(str)
+            else:
+                for str in [test.input, test.output]:
+                    all_strings.append(str)
+
+        # suggestions topics don't have topic markers so we check for them separately
+        all_strings.append("__suggestions__")
+        
+        # we don't use the output of the embedding, just do this to get the embeddings cached
+        adatest.embed(all_strings)
+
     def impute_labels(self):
         """ Impute missing labels in the test tree. """
         # TODO: this is just a random mock, it needs to implement real local topic models
 
-        self.compute_embeddings()
+        self._cache_embeddings()
 
         for id, test in self._tests.iterrows():
             if test.label == "":
-                features = np.hstack(adatest.embed([test.input, test.output]))
-                self._tests.loc[id, "label"] = self.topic_model(test.topic)(features)
+                if self.topic_membership_model(test.topic)(test.input) == "off_topic":
+                    self._tests.loc[id, "label"] = "off_topic"
+                else:
+                    self._tests.loc[id, "label"] = self.topic_labeling_model(test.topic)(test.input, test.output)
                 self._tests.loc[id, "labeler"] = "imputed"
 
-    def predict_labels(self, topical_io_pairs):
-        """ Return the label probabilities for a set of input-output pairs.
+    # def predict_labels(self, topical_io_pairs):
+    #     """ Return the label probabilities for a set of input-output pairs. [NOT USED RIGHT NOW]
 
-        Parameters
-        ----------
-        io_pairs : list[(str, str)]
-            A list of input-output pairs to score.
+    #     Parameters
+    #     ----------
+    #     io_pairs : list[(str, str)]
+    #         A list of input-output pairs to score.
 
-        Returns
-        -------
-        list[float]
-            A list of label probabilities.
-        """
+    #     Returns
+    #     -------
+    #     list[float]
+    #         A list of label probabilities.
+    #     """
 
-        out = np.zeros(len(topical_io_pairs))
+    #     out = np.zeros(len(topical_io_pairs))
 
-        to_embed = []
-        topics = {}
-        for i,(topic,input,output) in enumerate(topical_io_pairs):
-            if topic not in topics:
-                topics[topic] = []
-            to_embed.append(input)
-            to_embed.append(output)
-            topics[topic].append((i, len(to_embed) - 2, len(to_embed) - 1))
-        embeddings = adatest.embed(to_embed)
-        features = [None for i in range(len(topical_io_pairs))]
-        for topic in topics:
-            features = []
-            for i,ind1,ind2 in topics[topic]:
-                features.append(np.hstack([embeddings[ind1], embeddings[ind2]]))
-            features = np.vstack(features)
+    #     to_embed = []
+    #     topics = {}
+    #     for i,(topic,input,output) in enumerate(topical_io_pairs):
+    #         if topic not in topics:
+    #             topics[topic] = []
+    #         to_embed.append(input)
+    #         to_embed.append(output)
+    #         topics[topic].append((i, len(to_embed) - 2, len(to_embed) - 1))
+    #     embeddings = adatest.embed(to_embed)
+    #     features = [None for i in range(len(topical_io_pairs))]
+    #     for topic in topics:
+    #         features = []
+    #         for i,ind1,ind2 in topics[topic]:
+    #             features.append(np.hstack([embeddings[ind1], embeddings[ind2]]))
+    #         features = np.vstack(features)
 
-            label = np.array([v == "pass" for v in self.topic_model(topic)(features)], dtype=np.float32)
-            for i, (j,_,_) in enumerate(topics[topic]):
-                out[j] = label[i]
+    #         label = np.array([v == "pass" for v in self.topic_model(topic)(features)], dtype=np.float32)
+    #         for i, (j,_,_) in enumerate(topics[topic]):
+    #             out[j] = label[i]
 
-        return np.array(out)
+    #     return np.array(out)
 
+    def topic_labeling_model(self, topic):
+        topic = topic.replace("/__suggestions__", "") # predict suggestions using their parent topic label model
+        if topic not in self._topic_labeling_models:
+            self._topic_labeling_models[topic] = TopicLabelingModel(topic, self)
+        return self._topic_labeling_models[topic]
 
+    def topic_membership_model(self, topic):
+        topic = topic.replace("/__suggestions__", "") # predict suggestions using their parent topic membership model
+        if topic not in self._topic_membership_models:
+            self._topic_membership_models[topic] = TopicMembershipModel(topic, self)
+        return self._topic_membership_models[topic]
 
+    def retrain_topic_labeling_model(self, topic):
+        self._topic_labeling_models[topic] = TopicLabelingModel(topic, self)
 
-    def topic_model(self, topic):
-        if topic not in self._topic_models:
-            self._topic_models[topic] = TopicModel(topic, self)
-        return self._topic_models[topic]
-
-    def retrain_topic_model(self, topic):
-        self._topic_models[topic] = TopicModel(topic, self)
+    def retrain_topic_membership_model(self, topic):
+        self._topic_membership_models[topic] = TopicMembershipModel(topic, self)
 
     def drop_topic(self, topic):
         """ Remove a topic from the test tree. """
@@ -445,7 +447,7 @@ class TestTreeLocIndexer():
         
         subset = self.test_tree._tests.loc[key]
         if hasattr(subset, 'columns') and len(set(["topic", "input", "output", "label"]) - set(subset.columns)) == 0:
-            test_tree_slice = TestTree(subset)
+            test_tree_slice = TestTree(subset, index=subset.index, ensure_topic_markers=False)
             test_tree_slice._tests_location = self.test_tree._tests_location
             return test_tree_slice
         else:
@@ -467,7 +469,7 @@ class TestTreeILocIndexer():
         
         subset = self.test_tree._tests.iloc[key]
         if hasattr(subset, 'columns') and len(set(["topic", "input", "output", "label"]) - set(subset.columns)) == 0:
-            test_tree_slice = TestTree(subset)
+            test_tree_slice = TestTree(subset, ensure_topic_markers=False)
             test_tree_slice._tests_location = self.test_tree._tests_location
             return test_tree_slice
         else:
