@@ -4,7 +4,6 @@ import openai
 import numpy as np
 import copy
 import math
-import sentence_transformers
 import pandas as pd
 import torch
 import json
@@ -30,6 +29,7 @@ from ._scorer import expand_template, clean_template, ClassifierScorer, Generato
 from ._prompt_builder import PromptBuilder
 import builtins
 import adatest # Need to import like this to prevent circular dependencies
+import urllib.parse
 
 # from https://gist.github.com/walkermatt/2871026
 def throttle(interval):
@@ -72,6 +72,13 @@ def is_subtopic(topic, candidate):
     # Returns true if candidate is a subtopic of topic
     return True if re.search(r'^%s(/|$)' % topic.replace('+', r'\+'), candidate) else False
 
+def matches_filter(test, filter_text: str):
+    if filter_text is None or filter_text == "":
+        return True
+    else:
+        return filter_text in test["input"] or filter_text in test["output"]
+
+
 special_outputs = [
     "{MAX}"
 ]
@@ -116,6 +123,7 @@ class TestTreeBrowser():
         self.current_topic = starting_path
         self.score_filter = score_filter
         self.topic_model_scale = topic_model_scale
+        self.filter_text = ""
 
         # convert single generator to the multi-generator format
         if not isinstance(self.generators, dict):
@@ -180,7 +188,7 @@ class TestTreeBrowser():
         self.comm = None
 
         # define our current mode, and set of supported modes
-        self.mode = "tests"
+        self.mode = "tests" if self.test_tree.shape[0] > 0 else "topics"
         self.mode_options = [
             # "validity focused", # focus first on making valid in-topic tests, then secondarily on making those tests high scoring
             # "failure focused", # focus on making high scoring (failing) tests, then secondarily on making those tests valid and in-topic
@@ -189,8 +197,7 @@ class TestTreeBrowser():
         ]
 
         # apply all the scorers to the test tree (this updates the test tree)
-        for k in self.scorer:
-            self.scorer[k](self.test_tree, k+" score")
+        self._compute_embeddings_and_scores(self.test_tree, self.recompute_scores, overwrite_outputs=False, save_outputs=True)
 
         # # make sure all the tests have scores (if we have a scorer)
         # self._compute_embeddings_and_scores(self.test_tree)
@@ -198,10 +205,8 @@ class TestTreeBrowser():
         # ensure any test tree based generator has embeddings calculated
         if isinstance(self.generators, dict):
             for name, gen in self.generators.items():
-                if gen.gen_type == "test_tree":
-                    gen.source.compute_embeddings()
-        elif self.generators.gen_type == "test_tree": # HN: Probably unused code, should be safe to remove
-            self.generators.source.compute_embeddings()
+                if getattr(gen, "gen_type", "") == "test_tree":
+                    gen.source._cache_embeddings()
 
         # save the current state of the test tree
         self._auto_save()
@@ -228,7 +233,7 @@ class TestTreeBrowser():
 
         # dump the client javascript to the interface
         file_path = pathlib.Path(__file__).parent.absolute()
-        with open(file_path / ".." / "client" / "dist" / "main.js", encoding="utf-8") as f:
+        with open(file_path / "resources" / "main.js", encoding="utf-8") as f:
             js_data = f.read()
         interface_html = f"""
 <div id="adatest_container_{self._id}" style="width: 100%; all: initial;"></div>
@@ -274,7 +279,8 @@ class TestTreeBrowser():
                 # generate a new set of suggested tests/topics
                 elif action == "generate_suggestions":
                     self._clear_suggestions()
-                    self.test_tree.retrain_topic_model(self.current_topic)
+                    self.test_tree.retrain_topic_labeling_model(self.current_topic)
+                    self.test_tree.retrain_topic_membership_model(self.current_topic)
                     self._generate_suggestions(filter=msg[k].get("filter", ""))
                     # if self._active_generator_obj is None:
                     #     self._suggestions_error = "No AdaTest generator has been set!"
@@ -354,7 +360,7 @@ class TestTreeBrowser():
                         "description": ""
                     }
                     for c in self.score_columns:
-                        row[c] = ""
+                        row[c] = "__TOEVAL__"
                         row[c[:-6] + " raw outputs"] = "{}"
                     self.test_tree.loc[uuid.uuid4().hex] = row
 
@@ -384,12 +390,25 @@ class TestTreeBrowser():
                     self.active_generator = msg[k]["active_generator"]
                     self._active_generator_obj = self.generators[self.active_generator]
 
-                # change which generator is active
+                # change between topics and tests
                 elif action is None and "mode" in msg[k]:
                     self.mode = msg[k]["mode"]
 
                 elif action == 'change_description':
-                    self.test_tree.loc[msg[k]['topic_marker_id']]['description'] = msg[k]['description']
+                    id = msg[k]['topic_marker_id']
+                    if id not in self.test_tree.index:
+                        self.test_tree.loc[id, 'topic'] = "" # only the root topic would be missing from the tree
+                        self.test_tree.loc[id, 'input'] = ""
+                        self.test_tree.loc[id, 'output'] = ""
+                        self.test_tree.loc[id, 'label'] = "topic_marker"
+                    self.test_tree.loc[id, 'description'] = msg[k]['description']
+                    self._auto_save()
+
+                elif action == 'change_filter':
+                    print("change_filter")
+                    self.filter_text = msg[k]['filter_text']
+                    self._refresh_interface()
+
 
             # if we are just updating a single row in tests then we only recompute the scores
             elif "topic" not in msg[k]:
@@ -404,16 +423,22 @@ class TestTreeBrowser():
                 # update the row and recompute scores
                 for k2 in msg[k]:
                     self.test_tree.loc[k, k2] = msg[k][k2]
-                self.test_tree.loc[k, self.score_columns] = ""
-                    
-                self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs="input" in msg[k])
-                
+                if "input" in msg[k] or "output" in msg[k]:
+                    self.test_tree.loc[k, self.score_columns] = "__TOEVAL__"
+                    self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs="output" not in msg[k])
+                elif "label" in msg[k]:
+                    #self.test_tree.retrain_topic_model(self.current_topic)
+                    pass # SML: we could recompute the scores here but then that would change the output of stochastic output models
+                    # sign = -1 if msg[k]["label"] == "pass" else 1
+                    # self.test_tree.loc[k, self.score_columns] = ""#abs(float(self.test_tree.loc[k, self.score_columns])) #* sign
+                    # self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs=False)
 
                 # send just the data that changed back to the frontend
-                sendback_data["scores"] = {c: [[k, v] for v in score_parts(self.test_tree.loc[k, c])] for c in self.score_columns}
+                sendback_data["scores"] = {c: [[k, v] for v in ui_score_parts(self.test_tree.loc[k, c], self.test_tree.loc[k, "label"])] for c in self.score_columns}
                 outputs = {c: [[k, json.loads(self.test_tree.loc[k].get(c[:-6] + " raw outputs", "{}"))]] for c in self.score_columns}
                 sendback_data["raw_outputs"] = outputs
-                sendback_data["output"] = self.test_tree.loc[k, "output"]
+                if "output" not in msg[k]: # if the output was given to us the client is managing its current state so we shouldn't send it back
+                    sendback_data["output"] = self.test_tree.loc[k, "output"]
                 sendback_data["label"] = self.test_tree.loc[k, "label"]
                 sendback_data["labeler"] = self.test_tree.loc[k, "labeler"]
                 sendback_data.update(self.test_display_parts(self.test_tree.loc[k]))
@@ -426,7 +451,7 @@ class TestTreeBrowser():
                 
                 # move a test that is in the test tree
                 if k in self.test_tree.index:
-                    if msg[k]["topic"] == "_DELETE_" or msg[k]["topic"] == "_OUT_OF_TOPIC_": # this means delete the test
+                    if msg[k]["topic"] == "_DELETE_": # this means delete the test
                         self.test_tree.drop(k, inplace=True)
                     else:
                         self.test_tree.loc[k, "topic"] = msg[k]["topic"]
@@ -436,7 +461,7 @@ class TestTreeBrowser():
                 else:
                     for id, test in self.test_tree.iterrows():
                         if is_subtopic(k, test.topic):
-                            if msg[k]["topic"] == "_DELETE_" or msg[k]["topic"] == "_OUT_OF_TOPIC_":
+                            if msg[k]["topic"] == "_DELETE_":
                                 self.test_tree.drop(id, inplace=True)
                             else:
                                 self.test_tree.loc[id, "topic"] = msg[k]["topic"] + test.topic[len(k):]
@@ -480,14 +505,14 @@ class TestTreeBrowser():
                                 children.append(test.topic)
                     
                     # add a test
-                    else:
+                    elif matches_filter(test, self.filter_text):
                         data[k] = {
                             "input": test.input,
                             "output": test.output,
                             "label": test.label,
                             "labeler": test.labeler,
                             "description": test.description,
-                            "scores": {c: [[k, v] for v in score_parts(test[c])] for c in self.score_columns},
+                            "scores": {c: [[k, v] for v in ui_score_parts(test[c], test.label)] for c in self.score_columns},
                             "editing": test.input == "New test"
                         }
 
@@ -502,17 +527,18 @@ class TestTreeBrowser():
                     child_topic = test.topic[len(topic):].split("/", 2)[1]
                     scores = data[topic+"/"+child_topic]["scores"]
                     for c in self.score_columns:
-                        scores[c].extend([[k, v] for v in score_parts(test[c])])
+                        scores[c].extend([[k, v] for v in ui_score_parts(test[c], test.label)])
 
             # sort by score and always put new topics first
             def sort_key(id):
                 try:
                     total = 0
                     count = 0
+                    # offset = 0 if data[id]["label"] == "fail" else -1
                     for s in data[id]["scores"][self.score_columns[0]]:
                         val = score_max(s[1], nan_val=np.nan)
                         if not np.isnan(val) and val is not None:
-                            total += val
+                            total += val #+ offset
                             count += 1
                     if count == 0:
                         return 1e3
@@ -523,7 +549,9 @@ class TestTreeBrowser():
                     print(id)
                     print(val)
             sorted_children = sorted(children, key=sort_key)
-            sorted_children = sorted(sorted_children, key=lambda id: 0 if id.endswith("/New topic") or data[id].get("value1", "") == "New test" else 1)
+            sorted_children = sorted(sorted_children, key=lambda id: 0 if data[id].get("label", "") == "topic_marker" else 1) # put folders first
+            sorted_children = sorted(sorted_children, key=lambda id: 1 if data[id].get("label", "") == "off_topic" else 0) # off topic last
+            sorted_children = sorted(sorted_children, key=lambda id: 0 if id.endswith("/New topic") or data[id].get("value1", "") == "New test" else 1) # put new items first
 
             return sorted_children
         
@@ -585,6 +613,13 @@ class TestTreeBrowser():
             if self.test_tree.loc[k, "topic"].startswith(self.current_topic + "/__suggestions__"):
                 self.test_tree.drop(k, inplace=True)
 
+    def generate_suggestions(self, topic=None, filter=""):
+        if topic is not None:
+            self.current_topic = topic
+        self._clear_suggestions()
+        self.test_tree.retrain_topic_labeling_model(self.current_topic)
+        self.test_tree.retrain_topic_membership_model(self.current_topic)
+        self._generate_suggestions(filter=filter)
 
     def _generate_suggestions(self, filter):
         """ Generate suggestions for the current topic.
@@ -605,39 +640,6 @@ class TestTreeBrowser():
             else:
                 test_map[test.topic + " __JOIN__ " + test.input] = True
 
-        # see if we have a finite set of valid outputs
-        # valid_outputs = getattr(self.scorer, "output_names", None)
-        # if valid_outputs is not None and value2_filter is not None:
-        #     valid_outputs = [s for s in valid_outputs if re.search(value2_filter, s) is not None]
-
-        # # see if we have only topics are direct children, if so, we suggest topics
-        # has_direct_tests = False
-        # has_known_subtopics = False
-        # for k, test in self.test_tree.iterrows():
-        #     if test["topic"] == self.current_topic:
-        #         if test["type"] == "test":
-        #             has_direct_tests = True
-        #     elif is_subtopic(self.current_topic, test["topic"]):
-        #         has_known_subtopics = True
-        # suggest_topics = not has_direct_tests and has_known_subtopics
-        # suggest_topics = self.mode == "topics"
-        # zero_shot_tests = not has_direct_tests and not has_known_subtopics
-        
-        
-        
-        # {} should output {}
-        # {} should not output {}
-        # {} should have the same output as {}
-        # {} should not be more {} than {}
-        # {} should be invertable. [] is the output.
-        # {} should not output text containing {}
-        # {} should output toxic text. [] is the output.
-        # {} should not output toxic text. [] is the output.
-        
-        # {}'s output should not contain {}
-        
-
-
         
         # validity focused (focus first on making valid in-topic tests, then secondarily on making those tests high scoring)
         # failure focused (focus on making high scoring (failing) tests, then secondarily on making those tests valid and in-topic)
@@ -660,14 +662,31 @@ class TestTreeBrowser():
             suggest_topics=self.mode == "topics"
         )
 
+        # get the current topic description
+        curr_topic_mask = (self.test_tree["topic"] == self.current_topic) & (self.test_tree["label"] == "topic_marker")
+        if curr_topic_mask.sum() == 0:
+            desc = ""
+        else:
+            desc = self.test_tree.loc[(self.test_tree["topic"] == self.current_topic) & (self.test_tree["label"] == "topic_marker")]["description"][0]
+
         # generate the suggestions
-        proposals = self._active_generator_obj(prompts, self.current_topic, self.mode, self.scorer, num_samples=self.max_suggestions // len(prompts))
+        generators = [self._active_generator_obj] + list(self.generators.values())
+        for generator in generators:
+            try:
+                proposals = generator(prompts, self.current_topic, desc, self.mode, self.scorer, num_samples=self.max_suggestions // len(prompts) if len(prompts) > 0 else self.max_suggestions)
+                break
+            except ValueError:
+                pass # try the next generator
+        
+        # all topics should be URI encoded
+        if self.mode == "topics":
+            proposals = [urllib.parse.quote(x) for x in proposals]
         
         # Build up suggestions catalog, unless generating from a test tree source.
         # NOTE: Doing safe checks for TestTree type in order to prevent circular imports
         if isinstance(proposals, pd.DataFrame) or proposals.__class__.__name__ == "TestTree":
             suggestions = proposals
-            suggestions['topic'] = self.current_topic + "/__suggestions__" + suggestions['topic'].apply(lambda x: x[len(self.current_topic):] if x != "" else "/")
+            suggestions['topic'] = self.current_topic + "/__suggestions__" + suggestions['topic'].apply(lambda x: x[len(self.current_topic):] if x != "" else "")
             self.test_tree.append(suggestions)
             print("appended suggestions into self.test_tree")
             # assert False, "This needs to be fixed to dump into /__suggestions__"
@@ -685,12 +704,12 @@ class TestTreeBrowser():
                     id = uuid.uuid4().hex
                     self.test_tree.loc[id, "topic"] = self.current_topic + "/__suggestions__" + ("/"+input if self.mode == "topics" else "")
                     self.test_tree.loc[id, "input"] = "" if self.mode == "topics" else input
-                    self.test_tree.loc[id, "output"] = ""
+                    self.test_tree.loc[id, "output"] = "__TOOVERWRITE__"
                     self.test_tree.loc[id, "label"] = "topic_marker" if self.mode == "topics" else ""
                     self.test_tree.loc[id, "labeler"] = "imputed"
                     self.test_tree.loc[id, "description"] = ""
                     for c in self.score_columns:
-                        self.test_tree.loc[id, c] = ""
+                        self.test_tree.loc[id, c] = "__TOEVAL__"
 
                     # s = {
                     #     "topic": self.current_topic + "/__suggestions__" + ("/"+input if self.mode == "topics" else ""),
@@ -765,45 +784,48 @@ class TestTreeBrowser():
         for i, (k, row) in enumerate(df.iterrows()):
             row["score"] = float(row["score"]) + topic_model_scale * sim_scores[i]
 
-    def _compute_embeddings_and_scores(self, tests, recompute=False, overwrite_outputs=False): # TODO: Rename/refactor/merge with _compute_scores?
+    def _compute_embeddings_and_scores(self, tests, recompute=False, overwrite_outputs=False, save_outputs=False): # TODO: Rename/refactor/merge with _compute_scores?
         log.debug(f"compute_embeddings_and_scores(tests=<DataFrame shape={tests.shape}>, recompute={recompute})")
 
         for k in self.scorer:
-            self.scorer[k](tests, k+" score", overwrite_outputs=overwrite_outputs)
-        return
-        for k in self.scorer:
+            # determine which rows we need to evaluate
+            eval_ids = []
+            for i, (id, test) in enumerate(tests.iterrows()):
+                if (recompute or test[k+" score"] == "__TOEVAL__") and test.label != "topic_marker" and test.label != "off_topic":
+                    eval_ids.append(id)
 
-            self.scorer[k](tests)
+            if len(eval_ids) > 0:
 
-            # run the model on all the rows without a score
-            new_ids = tests.index[tests[k+ " score"] == ""]
-            outputs = self.scorer[k].outputs(tests["input"][new_ids])
+                # run the scorer
+                new_outputs,scores = self.scorer[k](tests, eval_ids)
 
-        if self.scorer is not None:
-            self._compute_scores(tests, recompute=recompute)
+                # update the scores in the test tree
+                current_outputs = tests["output"]
+                for i,id in enumerate(eval_ids):
+                    # tests.loc[id, k+" score"] = scores[i]
 
-        # model outputs and embeddings
-        if adatest.embedding_model is not None:
-            new_embedding_ids = [k for k in tests.index if k not in adatest._embedding_cache]
-            if len(new_embedding_ids) > 0:
-                value1s = []
-                value2s = []
-                value3s = []
-                for k in new_embedding_ids:
-                    if tests.loc[k, "type"] == "topic_marker":
-                        parts = tests.loc[k, "topic"].rsplit("/", 1)
-                        value1s.append(parts[1] if len(parts) == 2 else "")
-                        value2s.append("")
-                        value3s.append("")
+                    if not overwrite_outputs and current_outputs.loc[id] != "__TOOVERWRITE__" and current_outputs.loc[id] != new_outputs[i]:
+
+                        # mark the current row as nan score (meaning the output does not match)
+                        tests.loc[id, k+" score"] = np.nan
+
+                        # add a new test where the model output does match if we are saving outputs
+                        if save_outputs:
+                            id_new = uuid.uuid4().hex
+                            tests.loc[id_new, "topic"] = tests.loc[id, "topic"]
+                            tests.loc[id_new, "input"] = tests.loc[id, "input"]
+                            tests.loc[id_new, "output"] = new_outputs[i]
+                            tests.loc[id_new, "labeler"] = "imputed"
+                            tests.loc[id_new, "label"] = ""
+                            tests.loc[id_new, k+" score"] = scores[i]
                     else:
-                        value1s.append(str(tests.loc[k, "value1"]))
-                        value2s.append(str(tests.loc[k, "value2"]))
-                        value3s.append(str(tests.loc[k, "value3"]))
-                new_value1_embeddings = adatest.embedding_model.encode(value1s, convert_to_tensor=True, show_progress_bar=False).cpu()
-                new_value2_embeddings = adatest.embedding_model.encode(value2s, convert_to_tensor=True, show_progress_bar=False).cpu()
-                new_value3_embeddings = adatest.embedding_model.encode(value3s, convert_to_tensor=True, show_progress_bar=False).cpu()
-                for i,k in enumerate(new_embedding_ids):
-                    adatest._embedding_cache[k] = np.hstack([new_value1_embeddings[i], new_value2_embeddings[i], new_value3_embeddings[i]])
+                        tests.loc[id, "output"] = new_outputs[i]
+                        tests.loc[id, k+" score"] = scores[i]
+
+        tests.deduplicate() # make sure any duplicates we may have introduced are removed
+
+        # reimpute missing labels
+        tests.impute_labels() # TODO: ensure this method caches the local models and only reimputes when needed for each topic
 
     def _compute_scores(self, tests, recompute):
         """ Use the scorer(s) to fill in scores in the passed TestTree.
@@ -943,8 +965,8 @@ class TestTreeBrowser():
     def test_display_parts(self, test):
         
         # # find which template instantiation has the highest score (and so should be displayed)
-        # score_parts = test[self.score_columns[0]].split("|")
-        # if len(score_parts) == 1:
+        # ui_score_parts = test[self.score_columns[0]].split("|")
+        # if len(ui_score_parts) == 1:
         #     max_score_ind = 0
         # else:
         #     max_score_ind = np.argmax([float(v) for v in test[self.score_columns[0]].split("|")])
@@ -1055,11 +1077,24 @@ def score_max(s, nan_val=-1e3):
     else:
         return np.max(s)
 
-def score_parts(s):
-    if isinstance(s, str):
-        return [convert_float(v) for v in s.split("|")]
+def ui_score_parts(s, label):
+    """ Split a score into its parts and encode the label into the sign.
+    
+    Note this encoding is just used for passing scores to the UI (scores are not signed in the TestTree).
+    """
+    offset = 0
+    if label == "pass":
+        sign = 1
+        offset = -1 - 1e-6
+    elif label == "fail":
+        sign = 1
+        offset = 1e-6 # just so we have a positive number to encode that this was a failure
     else:
-        return [s]
+        sign = np.nan
+    if isinstance(s, str):
+        return [np.clip(offset + convert_float(v)*sign, -1, 1) for v in s.split("|")]
+    else:
+        return [np.clip(offset + s*sign, -1, 1)]
 
 def convert_float(s):
     if s == "":
