@@ -1,54 +1,42 @@
 """ A set of generators for AdaTest.
 """
 import asyncio
-
-import uuid
 import aiohttp
 import transformers
 import openai
-import copy
-from ._filters import clean_string
-import sentence_transformers
-import pandas as pd
+from profanity import profanity
 import numpy as np
 import torch
+import os
 import adatest
-from pathlib import Path
+import spacy
+from ._embedding import cos_sim
+
+try:
+    import clip
+    import clip_retrieval.clip_client
+except ImportError:
+    pass
 
 
 class Generator():
     """ Abstract class for generators.
     """
-    
-    def __init__(self, source, sep=None, subsep=None, quote=None, filter_profanity=None):
-        """ Create a new generator with the given separators and max length.
+
+    def __init__(self, source):
+        """ Create a new generator from a given source.
         """
         self.source = source
-        self.sep = sep
-        self.subsep = subsep
-        self.quote = quote
-        self.filter_profanity = filter_profanity
 
-        # value1_format = '"{value1}"'
-        # value2_format = '"{value1}"'
-        # value1_value2_format = '"{value1}" {comparator} "{value2}"'
-        # value1_comparator_value2_format = '"{value1}" {comparator} "{value2}"'
-        # value1_comparator_format = '"{value1}" {comparator} "{value2}"'
-        # comparator_value2_format = '"{value1}" {comparator} "{value2}"'
-
-    def __call__(self, prompts, max_length=None):
-        """ This should be overridden by concrete subclasses.
-
-        Parameters:
-        -----------
-        prompts: list of tuples, or list of lists of tuples
+    def __call__(self, prompts, topic, topic_description, mode, scorer, num_samples, max_length):
+        """ Generate a set of prompts for a given topic.
         """
-        pass
-    
+        raise NotImplementedError()
+
     def _validate_prompts(self, prompts):
         """ Ensure the passed prompts are a list of prompt lists.
         """
-        if isinstance(prompts[0][0], str):
+        if len(prompts[0]) > 0 and isinstance(prompts[0][0], str):
             prompts = [prompts]
         
         # Split apart prompt IDs and prompts
@@ -61,6 +49,36 @@ class Generator():
             trimmed_prompts.append(prompt_without_id)
 
         return trimmed_prompts, prompt_ids
+
+
+class TextCompletionGenerator(Generator):
+    """ Abstract class for generators.
+    """
+    
+    def __init__(self, source, sep, subsep, quote, filter):
+        """ Create a new generator with the given separators and max length.
+        """
+        super().__init__(source)
+        self.sep = sep
+        self.subsep = subsep
+        self.quote = quote
+        self.filter = filter
+
+        # value1_format = '"{value1}"'
+        # value2_format = '"{value1}"'
+        # value1_value2_format = '"{value1}" {comparator} "{value2}"'
+        # value1_comparator_value2_format = '"{value1}" {comparator} "{value2}"'
+        # value1_comparator_format = '"{value1}" {comparator} "{value2}"'
+        # comparator_value2_format = '"{value1}" {comparator} "{value2}"'
+
+    def __call__(self, prompts, topic, topic_description, max_length=None):
+        """ This should be overridden by concrete subclasses.
+
+        Parameters:
+        -----------
+        prompts: list of tuples, or list of lists of tuples
+        """
+        pass
     
     def _varying_values(self, prompts, topic):
         """ Marks with values vary in the set of prompts.
@@ -78,25 +96,33 @@ class Generator():
             # gen_value3 = False
         return show_topics#, gen_value1, gen_value2, gen_value3
     
-    def _create_prompt_strings(self, prompts, topic):
+    def _create_prompt_strings(self, prompts, topic, content_type):
         """ Convert prompts that are lists of tuples into strings for the LM to complete.
         """
 
-        show_topics = self._varying_values(prompts, topic)
+        assert content_type in ["tests", "topics"], "Invalid mode: {}".format(content_type)
+
+        show_topics = self._varying_values(prompts, topic) or content_type == "topic"
 
         prompt_strings = []
         for prompt in prompts:
             prompt_string = ""
             for p_topic, input in prompt:
                 if show_topics:
-                    prompt_string += self.sep + p_topic + ":" + self.sep + self.quote
+                    if content_type == "tests":
+                        prompt_string += self.sep + p_topic + ":" + self.sep + self.quote
+                    elif content_type == "topics":
+                        prompt_string += "A subtopic of " + self.quote + p_topic + self.quote + " is " + self.quote
                 else:
                     prompt_string += self.quote
                 
                 prompt_string += input + self.quote
                 prompt_string += self.sep
             if show_topics:
-                prompt_strings.append(prompt_string + self.sep + topic + ":" + self.sep + self.quote)
+                if content_type == "tests":
+                    prompt_strings.append(prompt_string + self.sep + topic + ":" + self.sep + self.quote)
+                elif content_type == "topics":
+                    prompt_strings.append(prompt_string + "A subtopic of " + self.quote + topic + self.quote + " is " + self.quote)
             else:
                 prompt_strings.append(prompt_string + self.quote)
         return prompt_strings
@@ -111,8 +137,8 @@ class Generator():
         num_samples = len(suggestion_texts) // len(prompts)
         samples = []
         for i, suggestion_text in enumerate(suggestion_texts):
-            if self.filter_profanity:
-                suggestion_text = clean_string(suggestion_text)
+            if callable(self.filter):
+                suggestion_text = self.filter(suggestion_text)
             prompt_ind = i // num_samples
             # prompt = prompts[prompt_ind]
             samples.append(suggestion_text)
@@ -131,10 +157,10 @@ class Generator():
         return list(set(samples))
 
            
-class Transformers(Generator):
-    def __init__(self, model, tokenizer, sep="\n", subsep=" ", quote="\"", filter_profanity=True):
+class Transformers(TextCompletionGenerator):
+    def __init__(self, model, tokenizer, sep="\n", subsep=" ", quote="\"", filter=profanity.censor):
         # TODO [Harsha]: Add validation logic to make sure model is of supported type.
-        super().__init__(source=model, sep=sep, subsep=subsep, quote=quote)
+        super().__init__(model, sep, subsep, quote, filter)
         self.gen_type = "model"
         self.tokenizer = tokenizer
         self.device = self.source.device
@@ -156,9 +182,11 @@ class Transformers(Generator):
 
         self._sep_stopper = StopAtSequence(self.quote+self.sep, self.tokenizer)
     
-    def __call__(self, prompts, topic, test_type=None, scorer=None, num_samples=1, max_length=100):
+    def __call__(self, prompts, topic, topic_description, mode, scorer=None, num_samples=1, max_length=100):
         prompts, prompt_ids = self._validate_prompts(prompts)
-        prompt_strings = self._create_prompt_strings(prompts, topic)
+        if len(prompts) == 0:
+            raise ValueError("ValueError: Unable to generate suggestions from completely empty TestTree. Consider writing a few manual tests before generating suggestions.") 
+        prompt_strings = self._create_prompt_strings(prompts, topic, mode)
         
         # monkey-patch a method that prevents the use of past_key_values
         saved_func = self.source.prepare_inputs_for_generation
@@ -199,60 +227,38 @@ class Transformers(Generator):
         return self._parse_suggestion_texts(suggestion_texts, prompts)
 
 
-class OpenAI(Generator):
+class OpenAI(TextCompletionGenerator):
     """ Backend wrapper for the OpenAI API that exposes GPT-3.
     """
     
-    def __init__(self, models, api_key=None, sep="\n", subsep=" ", quote="\"", temperature=1.0, top_p=0.95, filter_profanity=True):
+    def __init__(self, model="curie", api_key=None, sep="\n", subsep=" ", quote="\"", temperature=1.0, top_p=0.95, filter=profanity.censor):
         # TODO [Harsha]: Add validation logic to make sure model is of supported type.
-        super().__init__(models, sep=sep, subsep=subsep, quote=quote, filter_profanity=filter_profanity)
+        super().__init__(model, sep, subsep, quote, filter)
         self.gen_type = "model"
         self.temperature = temperature
         self.top_p = top_p
         if api_key is not None:
             openai.api_key = api_key
 
-    def __call__(self, prompts, topic, test_type, scorer, num_samples=1, max_length=100):
+        # load a key by default if a standard file exists
+        elif openai.api_key is None:
+            key_path = os.path.expanduser("~/.openai_api_key")
+            if os.path.exists(key_path):
+                with open(key_path) as f:
+                    openai.api_key = f.read().strip()
+
+    def __call__(self, prompts, topic, topic_description, mode, scorer, num_samples=1, max_length=100):
+        if len(prompts[0]) == 0:
+            raise ValueError("ValueError: Unable to generate suggestions from completely empty TestTree. Consider writing a few manual tests before generating suggestions.") 
+
         prompts, prompt_ids = self._validate_prompts(prompts)
         # prompt_strings = self._create_prompt_strings(prompts, topic)
 
         # find out which values in the prompt have multiple values and so should be generated
         topics_vary = self._varying_values(prompts, topic)
 
-        # TODO: bias generation towards model failures
-        # custom generation process for the "should not output" test
-        # if test_type == "{} should not output {}":
-
-        #     # create prompts to generate the inputs to the model
-        #     input_prompt_strings, _, _, _ = self._create_prompt_strings(prompts, topic, prefix=[], generate=[1])
-
-        #     # call the OpenAI API to complete the input generation prompts
-        #     response = openai.Completion.create(
-        #         engine=self.model, prompt=input_prompt_strings, max_tokens=max_length,
-        #         temperature=self.temperature, top_p=self.top_p, n=num_samples, stop=self.quote
-        #     )
-        #     input_suggestions = [choice["text"] for choice in response["choices"]]
-
-        #     output_scores = scorer.model(input_suggestions)
-        #     per_completion_token_bias_values = self._compute_bias_values(output_scores, scorer.model.output_names)
-
-        #     # create prompts to generate the outputs to the model
-        #     output_prompt_strings, _, _, _ = self._create_prompt_strings(prompts, topic, prefix=[1], generate=[2])
-
-        #     # call the OpenAI API to complete the output generation prompts
-        #     response = openai.Completion.create(
-        #         engine=self.model, prompt=output_prompt_strings, max_tokens=max_length,
-        #         temperature=self.temperature, top_p=self.top_p, n=num_samples, stop=self.quote,
-        #         per_completion_token_bias_values=per_completion_token_bias_values
-        #     )
-        #     output_suggestions = [choice["text"] for choice in response["choices"]]
-
-        #     # then we build the final suggestions as the combination of the input and output suggestions
-
-
-
         # create prompts to generate the model input parameters of the tests
-        prompt_strings = self._create_prompt_strings(prompts, topic)
+        prompt_strings = self._create_prompt_strings(prompts, topic, mode)
         
         # call the OpenAI API to complete the prompts
         response = openai.Completion.create(
@@ -263,37 +269,22 @@ class OpenAI(Generator):
         
         return self._parse_suggestion_texts(suggestion_texts, prompts)
 
-        # if self._should_generate_outputs(test_type, scorer):
-        #     model_inputs = self._create_model_inputs(test_type, inputs_filled)
-        #     outputs = self._generate_outputs(model_inputs, scorer)
-        #     output_prompt_strings = self._create_output_prompt_strings(prompts, topic)
 
-        #     for i in range(len(outputs)):
-        #         logit_bias = {self._quote_token: 1} # we also upweight the quote token so we don't decrease the chance of ending the output
-        #         for id in self._tokenizer(outputs[i])["input_ids"]:
-        #             logit_bias[id] = logit_bias.get(id, 0) + 1
-        #             response = openai.Completion.create(
-        #                 engine=self.model, prompt=output_prompt_strings, max_tokens=max_length,
-        #                 temperature=self.temperature, top_p=self.top_p, n=num_samples, stop=self.quote, logit_bias=logit_bias
-        #             )
-        #             suggestion_texts = [choice["text"] for choice in response["choices"]]
-
-
-class AI21(Generator):
+class AI21(TextCompletionGenerator):
     """ Backend wrapper for the AI21 API.
     """
     
-    def __init__(self, model, api_key, sep="\n", subsep=" ", quote="\"", temperature=0.95, filter_profanity=True):
+    def __init__(self, model, api_key, sep="\n", subsep=" ", quote="\"", temperature=0.95, filter=profanity.censor):
         # TODO [Harsha]: Add validation logic to make sure model is of supported type.
-        super().__init__(model, sep=sep, subsep=subsep, quote=quote, filter_profanity=filter_profanity)
+        super().__init__(model, sep, subsep, quote, filter)
         self.gen_type = "model"
         self.api_key = api_key
         self.temperature = temperature
         self.event_loop = asyncio.get_event_loop()
     
-    def __call__(self, prompts, topic, test_type=None, scorer=None, num_samples=1, max_length=100):
+    def __call__(self, prompts, topic, topic_description, mode, scorer=None, num_samples=1, max_length=100):
         prompts, prompt_ids = self._validate_prompts(prompts)
-        prompt_strings = self._create_prompt_strings(prompts, topic)
+        prompt_strings = self._create_prompt_strings(prompts, topic, mode)
         
         # define an async call to the API
         async def http_call(prompt_string):
@@ -328,7 +319,11 @@ class TestTreeSource(Generator):
         self.gen_type = "test_tree"
         self.assistant_generator = assistant_generator
 
-    def __call__(self, prompts, topic, test_type=None, scorer=None, num_samples=1, max_length=100): # TODO: Unify all __call__ signatures
+    def __call__(self, prompts, topic, topic_description, test_type=None, scorer=None, num_samples=1, max_length=100): # TODO: Unify all __call__ signatures
+        if len(prompts) == 0:
+            # Randomly return instances without any prompts to go off of. TODO: Consider better alternatives like max-failure?
+            return self.source.iloc[np.random.choice(self.source.shape[0], size=min(50, self.source.shape[0]), replace=False)]
+
         prompts, prompt_ids = self._validate_prompts(prompts)
 
         # TODO: Currently only returns valid subtopics. Update to include similar topics based on embedding distance?
@@ -349,15 +344,90 @@ class TestTreeSource(Generator):
         max_suggestions = min(num_samples * len(prompts), len(data_embeddings))
         method = 'distance_to_avg'
         if method == 'avg_distance':
-            dist = sentence_transformers.util.pytorch_cos_sim(topic_embeddings, data_embeddings)
+            dist = cos_sim(topic_embeddings, data_embeddings)
             closest_indices = torch.topk(dist.mean(axis=0), k=max_suggestions).indices
             
         elif method == 'distance_to_avg':
             avg_topic_embedding = topic_embeddings.mean(axis=0)
 
-            distance = sentence_transformers.util.pytorch_cos_sim(avg_topic_embedding, data_embeddings)
+            distance = cos_sim(avg_topic_embedding, data_embeddings)
             closest_indices = torch.topk(distance, k=max_suggestions).indices
 
         output = self.source.iloc[np.array(closest_indices).squeeze()].copy()
         output['topic'] = topic
-        return output
\ No newline at end of file
+        return output
+
+
+class ClipRetrieval(Generator):
+    """ Backend wrapper for the ClipRetrieval package and API.
+    """
+    
+    def __init__(self, indice_name="laion5B", url="https://knn5.laion.ai/knn-service", use_safety_model=True, use_violence_detector=True, deduplicate=True):
+        """ Build a new ClipRetrieval generator client.
+        """
+        super().__init__(indice_name)
+
+        # load our CLIP embedding model
+        self.clip_model, self.clip_preprocess = clip.load("ViT-L/14", device="cpu", jit=True)
+
+        # build a ClipRetrieval client
+        self.client = clip_retrieval.clip_client.ClipClient(
+            url=url,
+            indice_name=indice_name,
+            aesthetic_weight=0,
+            modality=clip_retrieval.clip_client.Modality.IMAGE,
+            num_images=10, # this will get overwritten inside our __call__ method
+            use_safety_model=use_safety_model,
+            use_violence_detector=use_violence_detector,
+            deduplicate=deduplicate
+        )
+
+    def __call__(self, prompts, topic, topic_description, mode, scorer, num_samples=1, max_length=100):
+        """ Generate suggestions for the given topic and prompts.
+        """
+
+        # update the client to return the correct number of images
+        self.client.num_images = num_samples
+
+        # make sure we have valid prompts
+        prompts, prompt_ids = self._validate_prompts(prompts)
+
+        # if no topic description is provided, use the topic as the description
+        # TODO: perhaps we can improve the prompt format over just /topic/subtopic here?
+        if topic_description == "":
+            topic_description = topic 
+
+        # embed the text representation of the topic
+        description_emb = self.get_text_embedding(topic_description)
+
+        # embed the images in the prompts
+        suggestion_texts = []
+        for p in prompts:
+
+            # we filter out all out-of-topic prompts because, unlike for text completion generators
+            # we can't condition on the topic for ClipRetrieval
+            prompt_embeds = [self.get_text_embedding(v[1]) for v in p if v[0].startswith(topic)]
+
+            if len(prompt_embeds) == 0 and topic_description == "":
+                raise ValueError("ValueError: Unable to generate suggestions from completely empty TestTree and no topic description. Consider adding some topics (or a topic description) before generating test suggestions.")
+
+            # get a mean embedding for the search query
+            # TODO: check if we should use spherical averaging here
+            # TODO: check why the clip_retrieval API doesn't distinguish between text and image embeddings (and hence why we can average them in the same space)
+            # TODO: perhaps we should model the distribution of embeddings and add some noise to the mean to get more diversity?
+            mean_embedding = np.mean(np.vstack(prompt_embeds + [description_emb]), axis=0)
+            
+            # get the top-k images from the ClipRetrieval API
+            response = self.client.query(embedding_input=mean_embedding.tolist())
+            suggestion_texts.extend(["__IMAGE="+result["url"] for result in response])
+        
+        # return a unique list of suggestions
+        # TODO: perhaps we should ensure the images are unique in content and not just url
+        return list(set(suggestion_texts))
+    
+    def get_text_embedding(self, text):
+        with torch.no_grad():
+            text_emb = self.clip_model.encode_text(clip.tokenize([text], truncate=True).to("cpu"))
+            text_emb /= text_emb.norm(dim=-1, keepdim=True)
+            text_emb = text_emb.cpu().detach().numpy().astype("float32")[0]
+        return text_emb
\ No newline at end of file
